import pyflakes.api as api
import requests
from flask import Flask, request
import sys
import os 
if not (os.path.dirname(os.path.abspath(__file__)).startswith(sys.path[-1])):
    from reporter import Reporter
    import pytranslator
else:
    from .reporter import Reporter
    from . import pytranslator

def p():
  app = Flask(__name__)
  print("The plugin is decreapted. Please use the CLI alongside a Studio+VSCode sync plugin.")
  @app.route('/', methods=["GET", "POST"]) 
  def base_page():
    code = (request.data).decode()
    try:
      translator = pytranslator.Translator()
      lua_code = translator.translate(code)
    except Exception as e:
      return "CompileError!:"+str(e)

    return lua_code

  @app.route('/err', methods=["GET", "POST"]) 
  def debug():
    code = (request.data).decode()
    rep = Reporter()
    num = str(api.check(code, "roblox.py", rep))
    print(num)
    return rep.diagnostics

  @app.route("/lib", methods=["GET"]) 
  def library():
      translator = pytranslator.Translator()
      return translator.get_luainit([])
    
  app.run(
  host='0.0.0.0', 
  port=5555 
  )


import sys

import os 
if not (os.path.dirname(os.path.abspath(__file__)).startswith(sys.path[-1])):
    import pytranslator, parser, header
    from errormanager import *
    from configmanager import *
    from util import *
else:
    from . import pytranslator, parser, header
    from .errormanager import *
    from .configmanager import *
    from .util import *



def cppcompile(r, file, pluscount=False):
  if file.endswith(".cpp") and os.environ.get("ENABLE_BETA_RPYC") == True:
    # compile the file to a file with the same name and path but .lua
    try:
      newctranslator = parser.CodeConverter(file, getconfig("c", "dynamiclibpath", "None"))
      newctranslator.parse(
      os.path.join(r, file),
      # C not C++
      flags=[
        '-I%s' % inc for inc in []
      ] + [
        '-D%s' % define for define in []
      ] + [
        '-std=%s' % getconfig("cpp", "std", "c++20")
      ] + [
        '-stdlib=%s' % getconfig("cpp", "stdlib", "libc++")
      ]
      )
      path = os.path.join(r, file)  
      newctranslator.diagnostics(sys.stderr)
      relative_path = backwordreplace(path,".cpp", ".lua", 1)
      with open(relative_path, 'w') as out:
        newctranslator.output(relative_path, out)   
                  
        #print(colortext.green("Compiled "+os.path.join(r, file)))
      if pluscount:
        pluscount.update(1)
        pluscount.current += 1
        #global count
        #count += 1
    except Exception as e:
      if "To provide a path to libclang use Config.set_library_path() or Config.set_library_file()" in str(e):
        print(error("dylib not found, use `roblox-pyc config`, c++, dynamiclibpath, and set the path to the dynamic library."))
      print(error(f"Compile Error!\n\n "+str(e), f"{os.path.join(r, file)}"))
      debug("Compiler error "+str(e))
      if pluscount:
        #pluscount.error()
        pluscount.update(1)
        pluscount.current += 1
        #global count
        #count += 1
        
        return 0      
  else:
    print(error("C is not enabled, please enable it in the config."))
    if pluscount:
      #pluscount.error()
      pluscount.update(1)
      pluscount.current += 1
      #global count
      #count += 1
      return 0
def ccompile(r, file, pluscount=False):
  if file.endswith(".c") and os.environ.get("ENABLE_BETA_RPYC") == True:
    # compile the file to a file with the same name and path but .lua
    try:
      newctranslator = parser.CodeConverter(file, getconfig("c", "dynamiclibpath", "None"))
      newctranslator.parse(
      os.path.join(r, file),
      # C not C++
      flags=[
        '-I%s' % inc for inc in []
      ] + [
        '-D%s' % define for define in []
      ] + [
        '-std=%s' % getconfig("c", "std", "c11")
      ] + [
        '-stdlib=%s' % getconfig("c", "stdlib", "libc")
      ]
      )
      path = os.path.join(r, file)  
                
      newctranslator.diagnostics(sys.stderr)
      relative_path = backwordreplace(path,".c", ".lua", 1)
      
      with open(relative_path, 'w') as out:
        newctranslator.output(relative_path, out)
                  
        #print(colortext.green("Compiled "+os.path.join(r, file)))
      if pluscount:
        pluscount.update(1)
        pluscount.current += 1
        #global count
        #count += 1
    except Exception as e:
      if "To provide a path to libclang use Config.set_library_path() or Config.set_library_file()" in str(e):
        print(error("dylib not found, use `roblox-pyc config`, c, dynamiclibpath, and set the path to the dynamic library."))
      print(error(f"Compile Error!\n\n "+str(e), f"{os.path.join(r, file)}"))
      debug("Compile error at "+str(e))
      if pluscount:
        #pluscount.error()
        pluscount.update(1)
        pluscount.current += 1
        #global count
        #count += 1
        return 0
  else:
    print(error("C is not enabled, please enable it in the config."))
    if pluscount:
      #pluscount.error()
      pluscount.update(1)
      pluscount.current += 1
      #global count
      #count += 1
      return 0
def pycompile(r, file, pluscount=False):
  # compile the file to a file with the same name and path but .lua
  contents = ""
            
  try:
    with open(os.path.join(r, file)) as rf:
      contents = rf.read()  
  except Exception as e:
    print(error(f"Failed to read {os.path.join(r, file)}!\n\n "+str(e)))
    # do not compile the file if it cannot be read
    return
            
  try:
    translator = pytranslator.Translator()
    lua_code = translator.translate(contents)
    #print(colortext.green("Compiled "+os.path.join(r, file)))
    # get the relative path of the file and replace .py with .lua
    path = os.path.join(r, file)  
      
    relative_path = backwordreplace(path,".py", ".lua", 1)
    
    if not os.path.exists(os.path.dirname(relative_path)):
      os.makedirs(os.path.dirname(relative_path))
    
    with open(relative_path, "w") as f:
      f.write(lua_code)
    
    if pluscount:
      #pluscount.error()
      pluscount.update(1)
      pluscount.current += 1
      #global count
      #count += 1
  except Exception as e:
    print(error(f"Compile Error!\n\n "+str(e), f"{os.path.join(r, file)}"))
    debug("Compile error at "+str(e))
    if pluscount:
      #pluscount.error()
      pluscount.update(1)
      pluscount.current += 1
      #global count
      #count += 1
      return 0
def lunarcompile(r, file, pluscount=False):
  # compile the file to a file with the same name and path but .lua
  # Run command and check if anything is outputted to stderr, stdout, or stdin
              
  process = subprocess.Popen(["moonc", os.path.join(r, file)], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  stdout, stderr = process.communicate()
              
  if stdout or stderr:
    if stdout:     
      print(error(f"Compile Error!\n\n "+str(stdout), f"{os.path.join(r, file)}"))
      if pluscount:
        #pluscount.error()
        pluscount.update(1)
        pluscount.current += 1
        #global count
        #count += 1
        return 0
    else:
      print(error(f"Compile Error!\n\n "+str(stderr), f"{os.path.join(r, file)}"))
      if pluscount:
        #pluscount.error()
        pluscount.update(1)
        pluscount.current += 1
        #global count
        #count += 1
        return 0
  else:
    try:
      newheader = header.lunarheader([])
                  
      # check if the new file has been created
      if os.path.exists(os.path.join(r, file.replace(".moon", ".lua"))):
        #print(colortext.green("Compiled "+os.path.join(r, file)))          
        with open(os.path.join(r, file.replace(".moon", ".lua")), "r") as f:
          contents = f.read()
        contents = backwordreplace(contents, "return", "", 1)
        
      else:
        print(error("File error for "+os.path.join(r, file)+"!"))
      if pluscount:
        pluscount.update(1)
        pluscount.current += 1
        #global count
        #count += 1
    except Exception as e:
        print(error(f"Compile Error!\n\n "+str(e), f"{os.path.join(r, file)}"))
        
        if pluscount:
          #pluscount.error()
          pluscount.update(1)
          pluscount.current += 1
          #global count
          #count += 1
          return 0
def robloxtscompile(r, file, pluscount=False):
  # Just add to pluscount, add later
  try:
    print(warn("At the moment roblox-ts is not supported, please wait for a future update."))
    if pluscount:
      #pluscount.error()
      pluscount.update(1)
      pluscount.current += 1
      #global count
      #count += 1
  except Exception as e:
      print(error(f"Compile Error!\n\n "+str(e), f"{os.path.join(r, file)}"))
      if pluscount:
        #pluscount.error()
        pluscount.update(1)
        pluscount.current += 1
        #global count
        #count += 1
        return 0
def allcompile(r, file, pluscount):
  # Checks what file it is then redirects to the correct compiler
  if file.endswith(".py"):
    pycompile(r, file, pluscount)
  elif file.endswith(".moon"):
    lunarcompile(r, file, pluscount)
  elif file.endswith(".c"): 
    ccompile(r, file, pluscount)
  elif file.endswith(".cpp"):
    cppcompile(r, file, pluscount)
  elif file.endswith(".ts"):
    robloxtscompile(r, file, pluscount)
  


"""Manages updating and installing dependencies"""

# PYPI 
from packaging import version

# FILES
import sys
import os 
if not (os.path.dirname(os.path.abspath(__file__)).startswith(sys.path[-1])):
    from errormanager import *
else:
    from .errormanager import * 

# BUILTIN
import subprocess,requests,pkg_resources, os

def check_luarocks():
    try:
        subprocess.call(["luarocks", "--version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)
        return True
    except FileNotFoundError:
        return False

def check_moonscript():
    try:
        subprocess.call(["moonc", "--version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)
        return True
    except FileNotFoundError:
        return False
      
def install_luarocks():
    print("Installing LuaRocks...")
    subprocess.call(["apt-get", "install", "-y", "luarocks"])

def install_moonscript():
    print("Installing MoonScript...")
    subprocess.call(["luarocks", "install", "moonscript", "--dev"])

def checkboth():
    if check_luarocks() == False:
        print(warn("LuaRocks is not installed, installing..."))
        install_luarocks()
    if check_moonscript() == False:
        print(warn("MoonScript is not installed, installing..."))
        install_moonscript()

def get_latest_version():
    url = f"https://pypi.org/pypi/roblox-pyc/json"
    response = requests.get(url)
    data = response.json()
    return data["info"]["version"]

def check_for_updates():
  current_version = pkg_resources.get_distribution("roblox-pyc").version
  latest_version = get_latest_version()
  if version.parse(latest_version) > version.parse(current_version):
    print(info(f"Update available to {latest_version}, you are currently using {current_version}"))
    choice = input("\t\tDo you want to update? (yes/no): ").lower()
    if choice == "yes":
      # Add the pip upgrade command here.
      # Get cfg.pkl 
      script_dir = os.path.dirname(os.path.realpath(__file__))
      returnval = ""
      try:
        with open(os.path.join(script_dir, "__communication__/cfg.pkl"), "rb") as file:
          returnval = file.read()
      except:
        print(error("Failed to safe-update, data is corrupted. Would you like to force-update, you may lose configuration data.", "auto-updater"))
        choice = input("\t\tDo you want to force-update? (yes/no): ").lower()
        if not choice == "yes":
          sys.exit()
      subprocess.run(["pip", "install", f"roblox-pyc=={latest_version}"])
      with open(os.path.join(script_dir, "__communication__/cfg.pkl"), "wb") as file:
        file.write(returnval)
        
        
def check_npm():
    try:
        subprocess.call(["npm", "--version"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL)
        return True
    except FileNotFoundError:
        return False
def install_npm():
    print("Installing NPM...")
    subprocess.call(["apt-get", "install", "-y", "npm"])




# CONFIG
# CFG will be in the same directory as this file
import pickle, os
cfgPath = os.path.join(os.path.dirname(os.path.realpath(__file__)), "__communication__/cfg.pkl")
def getconfig(arg1, arg2, default="None"):
    try:
      # Load the config file if it exists, or create a new one if it doesn't
      if os.path.exists(cfgPath):
          with open(cfgPath, "rb") as f:
              cfg = pickle.load(f)
      else:
          cfg = {}

      # Get the value from the config file, or use the default value if it doesn't exist
      value = cfg.get(arg1, {}).get(arg2, default)

      # Update the config file with the default value if it doesn't exist
      if arg1 not in cfg:
          cfg[arg1] = {}
      if arg2 not in cfg[arg1]:
          cfg[arg1][arg2] = default
          with open(cfgPath, "wb") as f:
              pickle.dump(cfg, f)

      return value
    except EOFError:
      # The file is empty, write a {} to it
      with open(cfgPath, "wb") as f:
        pickle.dump({}, f)
      return getconfig(arg1, arg2, default)

def setconfig(arg1, arg2, value, ignore=None):
    try:
      # Load the config file if it exists, or create a new one if it doesn't
      if os.path.exists(cfgPath):
          with open(cfgPath, "rb") as f:
              cfg = pickle.load(f)
      else:
          cfg = {}

      # Set the value in the config file and save it
      if arg1 not in cfg:
          cfg[arg1] = {}
      cfg[arg1][arg2] = value
      with open(cfgPath, "wb") as f:
          pickle.dump(cfg, f)
    except EOFError:
      # The file is empty, write a {} to it
      with open(cfgPath, "wb") as f:
        pickle.dump({}, f)
      return setconfig(arg1, arg2, value, ignore=ignore)



# TODO

import sys

import os 
if not (os.path.dirname(os.path.abspath(__file__)).startswith(sys.path[-1])):
    from loader import loader
    from textcompiler import *
    from configmanager import *
    from util import *
    from basecompilers import *
else:
    from .loader import loader
    from .textcompiler import *
    from .configmanager import *
    from .util import *
    from .basecompilers import *
import threading, shutil

def newIncli(basefile, func):
  def incli():
    # Get all the files inside of the path, look for all of them which are .py and even check inside of folders. If this is happening in the same directory as the script, do it in the sub directory test
    path = os.getcwd()

    #global count
    #count = 0
    localcount = 0
    for r, d, f in os.walk(path):
      for file in f:
        if basefile.__class__ == str:
          if file.endswith(basefile):
            localcount += 1
        else:
          for i in basefile:
            if file.endswith(i):
              localcount += 1
    newloader = loader(localcount)
    
    for r, d, f in os.walk(path):
      for file in f:
        if basefile.__class__ == str:
          if file.endswith(basefile):
            threading.Thread(target=func, args=(r, file, newloader)).start()
            #pycompile(r, file)
          else:
            othercompile(r, file)
        else:
          found = False
          for i in basefile:
            if file.endswith(i):
              threading.Thread(target=func, args=(r, file, newloader)).start()
              #pycompile(r, file)
              found = True
          if not found:
            othercompile(r, file)
    
    newloader.yielduntil()
    
    print(colortext.green("Compiled Files!"))
    if getconfig("general", "autocompile", False):
      action = input("")
      if action == "exit":
        exit(0)
      else:
        incli()
    else:
      exit(0)
  
  return incli

def newIncli2(basefile, func):
  def incli2():
    # Just like incli, but duplicates the direcotry with -compiled and compiles the files in there, also runs filtercompiledfolder() on the directory
    path = os.getcwd()+"-compiled"
    if os.path.exists(path):
      # delete the folder and child files
      shutil.rmtree(path)
    shutil.copytree(os.getcwd(), path)

    #global count
    #count = 0
    localcount = 0
    for r, d, f in os.walk(path):
      for file in f:
        # if basefile is a string just check for that, if it is a list, check for all of them
        if basefile.__class__ == str:
          if file.endswith(basefile):
            localcount += 1
        else:
          for i in basefile:
            if file.endswith(i):
              localcount += 1
          
    newloader = loader(localcount)
    
    for r, d, f in os.walk(path):
      for file in f:
        found = False
        for i in basefile:
          if file.endswith(i):
            threading.Thread(target=func, args=(r, file, newloader)).start()
            #pycompile(r, file)
            found = True
        if not found:
          othercompile(r, file)
    newloader.yielduntil()
    filtercompiledfolder()
    print(colortext.green("Compiled Files!"))
    if getconfig("general", "autocompile", False):
      action = input("")
      if action == "exit":
        exit(0)
      else:
        incli2()              
  return incli2

def newLanguage(basefile, func, commentart, p=None):
    def w():
      try:
        incli = newIncli(basefile, func)
        incli2 = newIncli2(basefile, func)
        
        if sys.argv.__len__() >= 1:
          if sys.argv[1] == "p":
            if p:
              p()
            else:
              print(error("Plugin is not supported for this language."))
          elif sys.argv[1] == "lib":
            # sys.argv[2] is the path to the file, create a new file there with the name robloxpyc.lua, and write the library to it
            lib()
          elif sys.argv[1] == "c":
            decreapted("c")
            # Go through every lua descendant file in the current directory and delete it and create a new file with the same name but .py
            confirm = input(warn("Are you sure? This will delete all .lua files and add a "+basefile+" file with the same name.\n\nType 'yes' to continue."))
            if confirm == "yes":   
              path = os.getcwd()
              
              for r, d, f in os.walk(path):
                for file in f:
                  if '.lua' in file:
                    luafilecontents = ""
                    with open(os.path.join(r, file), "r") as f:
                      luafilecontents = f.read()
                      
                    os.remove(os.path.join(r, file))
                    
                    # create new file with same name but  .py and write the lua file contents to it
                    open(os.path.join(r, file.replace(".lua", basefile)), "x").close()
                    # write the old file contents as a py comment
                    open(os.path.join(r, file.replace(".lua", basefile)), "w").write(commentart.format(luafilecontents))
                    print(colortext.green("Converted to py "+os.path.join(r, file)+" as "+file.replace(".lua", basefile)))
          elif sys.argv[1] == "cd":
            # Duplicate the cwd directory, the original will be renamed to -compiled and the new one will be renamed to the original. For the new one, go through every lua descendant file in the current directory and delete it and create a new file with the same name but .py
            confirm = input(warn("Are you sure? This will duplicate the current directory and compile the files in the new directory.\n\nType 'yes' to continue."))
            if confirm == "yes":
              path = os.getcwd()+"/src"
              # check if cwd/default.project.json exists, if it does, then use that as the default project file
              #if os.path.exists(os.getcwd()+"/default.project.json"):
              #  formatdefaultproj(os.getcwd()+"/default.project.json")
              # rename directory to -compiled
              os.rename(path, path+"-compiled")
              # duplicate the directory, remove the -compiled from the end
              shutil.copytree(path+"-compiled", path)
              # now we have 2 identical directories, one with -compiled and one without. for the one without, go through every lua descendant file in the current directory and delete it and create a new file with the same name but .py
              path = os.getcwd()

              for r, d, f in os.walk(backwordreplace(path, "-compiled", "", 1)):
                for file in f:
                  if '.lua' in file:
                    luafilecontents = ""
                    with open(os.path.join(r, file), "r") as f:
                      luafilecontents = f.read()
                      
                    os.remove(os.path.join(r, file))
                    
                    # create new file with same name but  .py and write the lua file contents to it
                    open(os.path.join(r, file.replace(".lua", basefile)), "x").close()
                    # write the old file contents as a C++ comment
                    open(os.path.join(r, file.replace(".lua", basefile)), "w").write("\"\"\"\n"+luafilecontents+"\n\"\"\"")
                    
                    print(colortext.green("Converted "+os.path.join(r, file)+" to "+file.replace(".lua", basefile)))
              # create a .rpyc file in the non -compiled directory
              open(os.path.join(backwordreplace(path, "-compiled", "", 1), ".rpyc"), "x").close()
              # set cwd to not -compiled
              os.chdir(backwordreplace(path, "-compiled", "", 1))
              print(info("Completed! You may need to modify the default.package.json or any other equivalent file to make it use the -compiled directory rather than the original."))
          elif sys.argv[1] == "w":
            print(colortext.magenta("Ready to compile ", os.path.join(os.path.dirname(os.path.realpath(__file__)))+" ...\n Type 'exit' to exit, Press enter to compile."))
            incli()
          elif sys.argv[1] == "d":
            print(colortext.magenta("Ready to compile ", os.path.join(os.path.dirname(os.path.realpath(__file__)))+" ...\n Type 'exit' to exit, Press enter to compile."))
            incli2()
          
      except IndexError:
        print(error("Invalid amount of arguments!", "roblox-py"))
      except KeyboardInterrupt:
        print(colortext.red("Aborted!"))
        sys.exit(0)
    return w


import json
import os
import sys
import os 
if not (os.path.dirname(os.path.abspath(__file__)).startswith(sys.path[-1])):
    from errormanager import *
    from util import *
else:
    from .errormanager import *
    from .util import *


def json_to_lua(json_str):
    data = json.loads(json_str)
    return _json_to_lua(data)

def _json_to_lua(data):
    if isinstance(data, dict):
        items = []
        for key, value in data.items():
            items.append('[{}] = {}'.format(_json_to_lua(key), _json_to_lua(value)))
        return '{{{}}}'.format(', '.join(items))
    elif isinstance(data, list):
        items = []
        for value in data:
            items.append(_json_to_lua(value))
        return '{{{}}}'.format(', '.join(items))
    elif isinstance(data, str):
        return '"{}"'.format(data)
    else:
        return str(data)


def unknowncompile(r, file):
  if file.endswith(".txt") or ("." not in file):
     # This is for text files and files without a file extension
    try:
      contents = ""
      with open(os.path.join(r, file), "r") as f:
        contents = f.read()
        contents = contents.replace("]]", "]\]")
        contents = "--/ Compiled using roblox-pyc | Textfile compiler \--\nlocal file\nfile = {Contents = [["+contents+"]], Type = 'rawtext', Extension = '"+file.split(".")[file.split(".").__len__()-1]+"', SetSource = function(self, input) self.Contents = input end}"
        filename = os.path.basename(file)
        sepratedbydot = filename.split(".")
        ending = sepratedbydot[sepratedbydot.__len__()-1]
        newfilename = filename.replace("."+ending, ".lua")
        # if newfilename == oldfilename, add .lua to the end. For files without endings
        if filename == newfilename:
          newfilename = newfilename+".lua"
                  
        open(os.path.join(r, newfilename), "x").close()
        with open(os.path.join(r, newfilename), "w") as f:
          f.write(contents)
    except UnicodeDecodeError:
        print(warn("Failed to read "+os.path.join(r, file)+"!"))
def jsoncompile(r, file):
  if file.endswith(".json"):
    # compile the file to a file with the same name and path but .lua
    try:
      contents = ""
      with open(os.path.join(r, file), "r") as f:
        contents = f.read()
        contents = "--/ Compiled using roblox-pyc | JSON compiler \--\nreturn {Contents = "+json_to_lua(contents)+", Type = 'json', Extension = 'json', SetSource = function(self, input) self.Contents = input end}"
        filename = os.path.basename(file)
        sepratedbydot = filename.split(".")
        ending = sepratedbydot[sepratedbydot.__len__()-1]
        newfilename = filename.replace("."+ending, ".lua")
        # if newfilename == oldfilename, add .lua to the end. For files without endings
        if filename == newfilename:
          newfilename = newfilename+".lua"
        if not os.path.exists(os.path.dirname(os.path.join(r, newfilename))):
          open(os.path.join(r, newfilename), "x").close()
        with open(os.path.join(r, newfilename), "w") as f:
          f.write(contents)
    except UnicodeDecodeError:
        # Just delete the file
        print(warn("Failed to read "+os.path.join(r, file)+"!"))
def othercompile(r, file): # Handles Text files and JSON files, and files without a file extension
  jsoncompile(r, file)
  unknowncompile(r, file)